from .config import __version__, LOG_LEVEL, LOG_FORMAT
from .singleton import Singleton
from .color_item import ColorItem
from .wheel_item import WheelItem
from .colorwheels_config import ColorwheelsConfig
from .colorwheels import Colorwheels



"""ColorwheelsConfig is a configuration helper for :doc:`colorwheels`, implemented
as a singleton.

The helper loads a configuration YAML file and serves the colors by name to
a Colorwheels generator."""

import logging

import yaml

from .config import LOG_LEVEL, LOG_FORMAT
from .singleton import Singleton
from .color_item import ColorItem
from .wheel_item import WheelItem

logger = logging.getLogger(__name__)
logging.basicConfig(format=LOG_FORMAT, level=LOG_LEVEL)

class ColorwheelsConfig(metaclass=Singleton):
    """ helper"""

    def __init__(self):
        """ Initialize configuration helper for :doc:`colorwheels`.

        The constructor creates the most simple configuration, and we expect
        a configuration file to be loaded later by code.

        The initial wheels available are primitives: 'red', 'green', 'blue'"""

        self._wheel_items = list() # list of WheelItem objects

        self.release = "unknown"

        self._wheel_items.append(
            WheelItem("default", [
                ColorItem(255, 0, 0),
                ColorItem(0, 255, 0),
                ColorItem(0, 0, 255)]))

# -- Configuration Handling --------------------------------------------------

    @property
    def first_wheel(self):
        """Get the first wheel available"""

        # should be always available.... no need to raise
        return self._wheel_items[0]

    def find_wheel(self, name):
        """Find wheelitem by name. None if not found"""

        items = [item for item in self._wheel_items if item.name == name]
        return items[0] if items else None

    def wheel_names(self):
        """Return list of wheel names available in configuration"""

        names = list()
        for wheel in self._wheel_items:
            names.append(wheel.name)
        return names

    def add_wheel_item(self, item):
        """Adds a :doc:`wheel_item` to definitions list.

        Raises
        ------
        ValueError
            Raises error if item name already exists
        """

        if not isinstance(item, WheelItem):
            raise ValueError("cannot add object. Not of type WheelItem")

        if self.find_wheel(item.name) is None:
            self._wheel_items.append(item)
        else:
            raise ValueError("Item '%s' cannot be added. Already exists" % item.name)

    def create_wheel_item(self, name, colors):
        """Create a :doc:`wheel_item` from parts. Function returns the created item"""

        new_item = WheelItem(name=name, colors=colors)
        return new_item

# -- Loading Configuration ---------------------------------------------------

    def _check_release(self):
        """Validate the version of the configuration file."""

        # First version doesn't need any validation. Reserved for future
        logger.info("Color definition file is at version '%s'", self.release)

    def _create_wheel_items(self, yaml_data):
        """Convert yaml data to wheel items"""

        self._wheel_items.clear()

        # handle metadata
        self.release = yaml_data["meta"]["release"]
        self._check_release()

        # handle wheel elements
        for wheel_def in yaml_data["wheels"]:

            color_list = list()

            name = wheel_def["wheel"]["name"]
            element_type = wheel_def["wheel"]["type"] \
                if "type" in wheel_def["wheel"] else "sequence"

            if element_type == "sequence":
                for color_def in wheel_def["wheel"]["colors"]:
                    col = ColorItem(color_def["rgb"][0], color_def["rgb"][1], color_def["rgb"][2])
                    color_list.append(col)

                new_def = WheelItem(name, color_list)
                self.add_wheel_item(new_def)

            elif element_type == "rainbow":
                new_def = WheelItem(name, color_list)
                # Read rainbow parameters and use sensible defaults if not available
                rainb_size = wheel_def["wheel"]["size"] \
                    if "size" in wheel_def["wheel"] else 32
                rainb_amplitude = wheel_def["wheel"]["amplitude"] \
                    if "amplitude" in wheel_def["wheel"] else 127
                rainb_center = wheel_def["wheel"]["center"] \
                    if "center" in wheel_def["wheel"] else 128
                rainb_frequency = wheel_def["wheel"]["frequency"] \
                    if "frequency" in wheel_def["wheel"] else 0.3
                new_def.generate_rainbow(rainb_size, rainb_amplitude, rainb_center, rainb_frequency)

                self.add_wheel_item(new_def)
            else:
                raise ValueError(f"Unknown wheel type {element_type}")

    def load_wheels(self, filename):
        """loads YAML color definition file. The loaded file is converted to a list of
        :doc:`wheel_item` objects.

        Raises ``FileNotFoundError``, or ``YAMLError`` if loading fails.

        When loading a new definition from YAML, make sure colorwheels activates whatever wheel
        is required!"""

        try:
            with open(filename, 'r') as stream:
                try:
                    yml = yaml.safe_load(stream)
                    self._create_wheel_items(yml)
                except yaml.YAMLError as exc:
                    logger.fatal("Loading configuration file '%s' failed: %s", filename, exc)
                    raise exc

        except FileNotFoundError as exc:
            logger.fatal("File '%s' not found ...", filename)
            raise exc



# config.py

import logging

__version__ = "0.7.1"

#: Sets application-wide loglevel
LOG_LEVEL = logging.DEBUG
#: Set logger message format
LOG_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'



#!/usr/bin/env python

""" ``Colorwheels`` is a module to generate color sequences. A Colorwheels instance
contains multiple color wheel definitions and functions as a generator of color values
in multiple formats. You can activate any available colorwheel anytime, to obtain
different effects.
"""

import logging

from .config import LOG_LEVEL, LOG_FORMAT
from .colorwheels_config import ColorwheelsConfig

logger = logging.getLogger(__name__)
logging.basicConfig(format=LOG_FORMAT, level=LOG_LEVEL)

generator_types = ["rgb_tuple", "rgba_tuple", "hexadecimal"]

class Colorwheels():
    """Base class for returning color sequences.

    We represent colors similar to Color Wheels in photography, i.e. a sequence
    of colors is located on an imaginary wheel and endlessly served to applications
    in a given sequence.

    Colorwheels can be generated or loaded from a YAML file. See documentation and
    examples.
    """

    def __init__(self):
        """Create ``Colorwheels`` instance.

        wheel configurations is a :doc:`colorwheels_config` object (singleton), which
        contains color definitions loaded from the environment, or generated
        by code.

        A default configuration object is created (or inherited from other
        parts of the code).
        """

        self.counter = 0
        self._generator_type = ""
        self.set_generator_type("rgb_tuple")

        self._wheel_configurations = ColorwheelsConfig()

        self.active_wheel = None
        self.activate_colorwheel(self._wheel_configurations.first_wheel.name)

    def __str__(self):
        return (f"Colorwheels - wheels: {len(self._wheel_configurations)},"
                f"active: '{self.active_wheel.name}'")

    @property
    def wheel_configurations(self):
        """Returns the configuration object :doc:`colorwheels_config`,
        used by Colorwheels

        You can manage the configurations from any Colorwheels instance used
        in your program. Handle with care! The configurator is a singleton, i.e.
        if you for example load a different set of colors, all running
        generators will be affected.
        """

        return self._wheel_configurations

# -- Generator Functions -----------------------------------------------------

    def __iter__(self):
        """Return the iterator"""

        return self

    def __next__(self):
        """Return next color element. The return value depends on
        generator_type"""

        if self._generator_type == "rgb_tuple":
            return self.next()
        if self._generator_type == "rgba_tuple":
            return self.next_rgba()
        if self._generator_type == "hexadecimal":
            return self.next_hex()

        raise ValueError(f"Unknown generator type '{self._generator_type}'")

    def _next_color_item(self):
        """Returns a next ColorItem available in active ColorWheel"""

        # this is the core method of all iterator flavors below
        color = self.active_wheel.colors[self.counter]
        self.counter += 1
        if self.counter >= len(self.active_wheel.colors):
            self.counter = 0
        return color

    def next(self):
        """Get the next color from the ColorWheel as an RGB tuple

        Returns
        -------
        tuple
            RGB tuple of next selected color. The returned value is an integer
            tuple representing a color, i.e. (r,g,b). Red would return (255,0,0)

        See Also
        --------
        next_rgba: Get the next color from ColorWheel using RGBA
        next_hex: Get the next color from ColorWheel as a hex string
        """

        color = self._next_color_item()
        return (color.red, color.green, color.blue)

    def next_rgba(self, alpha=255):
        """Get the next color from ColorWheel using RGBA

        Returns
        -------
        tuple
            RGB tuple of next selected color. The returned value is an integer
            tuple representing a color, i.e. (r,g,b,a). Red would return
            (255,0,0,255) if alpha is default at 255

        See Also
        --------
        next: Get the next color from the ColorWheel as an RGB tuple
        next_hex: Get the next color from ColorWheel as a hex string
        """

        color = self.next()
        color = color + (alpha,)
        return color

    def next_hex(self):
        """Get the next color from ColorWheel as a hex string

        Returns
        -------
        string
            hex string representation of RGB color

        See Also
        --------
        next: Get the next color from the ColorWheel as an RGB tuple
        next_rgba: Get the next color from ColorWheel using RGBA
        """

        color = self.next()
        return "#{:02x}{:02x}{:02x}".format(color[0], color[1], color[2])

    def set_generator_type(self, new_type):
        """
        Set the generator type to a value out of generator_types.

        Raises
        ------
        ValueError
            Raises ValueError exception if new_type is not avaialble
        """

        if new_type not in generator_types:
            raise ValueError(f"Unknown generator type '{new_type}'")

        if new_type != self._generator_type:
            self._generator_type = new_type
            logger.info("Setting generator type to '%s'", self._generator_type)

# -- NEW ---------------------------------------------------------------------

    def activate_colorwheel(self, name):
        """Activates colorwheel by name, from configuration file. Sets
        active_wheel with new setting.

        Raises
        ------
        ValueError
            Raises ValueError exception if name is not found
        """

        self.counter = 0 # always reset counter
        self.active_wheel = self._wheel_configurations.find_wheel(name)

        if self.active_wheel is not None:
            logger.info("Activating wheel '%s'", name)
        else:
            message = f"Wheel '{name}' cannot be activated. Not found"
            logger.fatal(message)
            raise ValueError(message)



"""ColorItem is a dataclass containing one Color definition and color converters.

The object contains the following values:

* red: red component of color (integer)
* green: green component of color (integer)
* blue: blue component of color (integer)

You can convert color formats as well in between RGB, RGBA (integer representation)
and float representations.
"""

from dataclasses import dataclass

@dataclass
class ColorItem:
    """Color object (dataclass) for easy color handling.

    Parameters
    ----------
    red
        red color component. The native format is an integer 0-255
    green
        green color component. The native format is an integer 0-255
    blue
        blue color component. The native format is an integer 0-255
    """

    red: int
    green: int
    blue: int

    @property
    def color(self):
        """The `color` property return color as tuple.

        Returns
        -------
        tuple
            A tuple of 3 elements, red, green, blue. A red is returned
            as `(255, 0, 0)`
        """

        return (self.red, self.green, self.blue)

    @property
    def color_hex(self):
        """Return hexadecimal representation of color.

        Returns
        -------
        str
            Hexadecimal coded string. A red is returned as `#ff0000`
        """

        return "#{:02x}{:02x}{:02x}".format(self.red, self.green, self.blue)

    @property
    def complement(self):
        """Return complement (opposite) color tuple.

        Returns
        -------
        tuple
            Finds a complementing color to current, and returns it as a tuple.
            Our example red color - `(255, 0, 0)` is complemented by
            `(0, 255, 255)` - cyan.
        """

        def min_max(a, b, c):
            """inner function to find best delta for complementing color"""

            if c < b:
                b, c = c, b
            if b < a:
                a, b = b, a
            if c < b:
                b, c = c, b
            return a + c

        k = min_max(self.red, self.green, self.blue)
        return tuple(k - u for u in (self.red, self.green, self.blue))

    def from_float(self, colrgb):
        """Convert a float RGB tuple the native format (int tuple)

        This method comes in handy, if you use libraries like 'Colour' in your
        code.

        The Colour library uses RGB float values, encoded in tuples ranging
        from 0.0-1.0. We convert these values to an int tuple, i.e. int values
        ranging from 0-255
        """

        self.red = int(255*colrgb[0])
        self.green = int(255*colrgb[1])
        self.blue = int(255*colrgb[2])



"""colorwheels setuptools setup module
"""

# description of options of setup.py in https://github.com/pypa/sampleproject

from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

# read config.py settings, including version
exec(open(here / 'src/colorwheels/config.py').read())

setup(
    name='colorwheels',
    version=__version__,
    description='colorwheels - an endless color generator',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/quantumspaces/colorwheels',
    author='Quantum Spaces',
    author_email='quantumspaces@outlook.com',

    classifiers=[ # List of valid classifiers: https://pypi.org/classifiers/
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
    ],

    keywords='development, color',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    python_requires='>=3.7, <4',

    install_requires = [
        'dataclasses-json',
        'PyYAML'
    ],

    extras_require={  # Optional
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },

    project_urls={
        'Bug Reports': 'https://github.com/quantumspaces/colorwheels/issues',
        'Source': 'https://github.com/quantumspaces/colorwheels',
    },
)



"""WheelItem: a dataclass containing one ColorWheel definition

The object contains the following values:

* name: ColorWheel name. The name is used to retrieve a named color sequence
* colors: List of ColorItem colors

WheelItem encapsulates a colorwheel selection and is used internally by
colorwheels.
"""

import math
from dataclasses import dataclass
from typing import List

from .color_item import ColorItem

@dataclass
class WheelItem:
    """Content of one colorwheel"""

    name: str
    colors: List[ColorItem]

    def generate_rainbow(self, size, amplitude, center, frequency):
        """Generate colors with a Rainbow palette. Overwrites colors list.

        Uses a simplified algorithm.

        Tip: If you don't wish to experiment with the algorithm, you can
        decide on the rainbow size (number of colors), and use these values
        as a starting point:

        amplitude=127, center=128, frequency=0.3
        """

        self.colors.clear()

        for i in range(size):
            red = math.sin(frequency*i + 0) * amplitude + center
            green = math.sin(frequency*i + 2) * amplitude + center
            blue = math.sin(frequency*i + 4) * amplitude + center
            self.colors.append(ColorItem(red=int(red), green=int(green), blue=int(blue)))

    def from_float_list(self, color_list):
        """Convert a list of float RGB tuples to native format

        This method comes in handy, if you use libraries like 'Colour' in your
        code.

        The Colour library uses RGB float values, encoded in tuples ranging
        from 0.0-1.0. We convert these values to an int tuple, i.e. int values
        ranging from 0-255
        """

        self.colors.clear()

        for color in color_list:
            new_color = ColorItem(0, 0, 0)
            new_color.from_float(color)
            self.colors.append(new_color)



# singleton.py

"""Definition of a Singleton Meta-class. A class which attaches this metaclass
starts behaving as a singleton instance.

Borrowed from: https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
"""

#using in Python 2
# class MyClass(BaseClass):
#     __metaclass__ = Singleton

#using in Python 3
# class MyClass(BaseClass, metaclass=Singleton):
#     pass

class Singleton(type):
    """Singleton metaclass"""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]
